import typing

from .cycle_prevention import CyclePrevention
from .filefinder import FileFinder
from .rules import (
    BASIC_INCLUDE_RULES,
    BASIC_SKIP_RULES,
    Import,
    apply_skip_rules,
    apply_subimport_substitution_rules,
    apply_substitution_rules,
    find_imports,
)
from .traceable_string import TraceableString


class Preprocessor:
    def __init__(self, project_root: str):
        """
        :param project_root: The root of the LaTeX-document. Important for resolving
        the relative paths of include statements.
        """
        self.skip_rules = list(BASIC_SKIP_RULES)
        self.subimport_rules = []
        self.substitution_rules = []
        self.import_rules = list(BASIC_INCLUDE_RULES)
        self.file_finder = FileFinder(project_root)
        self.structure = {}

    def read_file(self, file_path: str) -> TraceableString:
        """
        Read a file and apply the SkipRules and SubstitutionRules.
        :param file_path: Path to the file.
        :return: Preprocessed file content
        """
        content = TraceableString(self.file_finder.read(file_path), origin=file_path)
        content = apply_skip_rules(content, self.skip_rules)
        content = apply_substitution_rules(content, self.substitution_rules)
        return content

    def include_path(
        self, content: TraceableString, subimport_path: str
    ) -> TraceableString:
        if subimport_path is None or subimport_path == "":
            return content
        content = apply_subimport_substitution_rules(
            content, self.subimport_rules, subimport_path
        )
        return content

    def find_imports(self, content: TraceableString) -> typing.List[Import]:
        """
        Find all imports within the file content (based on the import rules).
        :param content: The content of the LaTeX-file.
        :return: List of imports in the content.
        """
        imports = find_imports(content, self.import_rules)
        return imports

    def _add_structure(self, path: str, included_files: typing.List[str]):
        self.structure[path] = {
            "content": self.file_finder.read(path),
            "includes": included_files,
        }

    def expand_file(
        self,
        file_path: str,
        _cycle_prevention: typing.Optional[CyclePrevention] = None,
        is_subimport: bool = False,
        subimport_path: str = None,
    ) -> TraceableString:
        """
        Expand/flatten the file. This is performed recursively, but there will be an
        excepetion in case of cyclic include-commands.
        :param file_path: The path to the file to be included.
        :param _cycle_prevention: Internal use for preventing cyclic inclusions.
        :return: A flat LaTeX-document containing all included files.
        """
        _cycle_prevention = (
            _cycle_prevention if _cycle_prevention else CyclePrevention()
        )
        offset = 0
        _cycle_prevention.push(file_path, context=file_path)
        content = self.read_file(file_path)
        imports = self.find_imports(content)
        self._add_structure(file_path, [import_.path for import_ in imports])
        for match in imports:
            insertion_file = self.file_finder.find_best_matching_path(
                match.path, origin=file_path
            )
            insertion = self.expand_file(
                insertion_file,
                _cycle_prevention,
                match.is_subimport,
                match.subimport_path,
            )
            content = (
                content[: match.start + offset]
                + insertion
                + content[match.end + offset :]
            )
            offset += len(insertion) - len(match)
        if is_subimport:
            content = self.include_path(content, subimport_path)
        _cycle_prevention.pop()
        return content



# flake8: noqa F401
from .import_rules import (
    NativeImportRule,
    SubimportRule,
    ExplicitImportRule,
    Import,
    ImportRule,
    find_imports,
)
from .skip_rules import BasicSkipRule, SkipRule, TodonotesRule, apply_skip_rules
from .substitution_rules import (
    ChangesRule,
    SubstitutionRule,
    Substitution,
    apply_substitution_rules,
)
from .subimport_substiution_rules import (
    SubimportChangesRule,
    SubimportSubstitutionRule,
    SubimportSubstitution,
    apply_subimport_substitution_rules,
)

BASIC_INCLUDE_RULES = [NativeImportRule(), SubimportRule(), ExplicitImportRule()]
BASIC_SKIP_RULES = [BasicSkipRule()]

__ALL__ = [
    "BASIC_INCLUDE_RULES",
    "BASIC_SKIP_RULES",
    "NativeImportRule",
    "SubimportRule",
    "ExplicitImportRule",
    "Import",
    "ImportRule",
    "find_imports",
    "BasicSkipRule",
    "SkipRule",
    "TodonotesRule",
    "apply_skip_rules",
    "ChangesRule",
    "SubstitutionRule",
    "Substitution",
    "apply_substitution_rules",
    "SubimportChangesRule",
    "SubimportSubstitutionRule",
    "SubimportSubstitution",
    "apply_subimport_substitution_rules",
]



"""
Parsing LaTeX-commands perfectly is difficult, because the rules are quite complex.
There are many attempts using regular expressions, but I haven't seen one that
could not be tricked.
The following code can still be tricked by implicit parameters (no brackets {}),
but otherwise should be safe.
"""

import logging
import typing

_logger = logging.getLogger(__file__)


class _ParserError(ValueError):
    def __init__(self, msg, position):
        super().__init__(msg)
        self.position = position


class LatexStream:
    """
    The LatexStreams job is to keep track of the reading position and
    whether we are currently in a comment section or the current symbol
    is escaped.
    """

    def __init__(self, text: str, pos: int = 0):
        self._text = text
        self._pos = pos
        self.is_escaped = False
        self.in_comment = False
        self._read_escape = False

    def next(self) -> str:
        """
        Return the next character and move the cursor the next position.
        Update the current status regarding comment or escaping.
        :return: Next character.
        """
        if not self.has_next():
            msg = "No next character."
            raise _ParserError(msg, self._pos)
        c = self._text[self._pos]
        self._pos += 1
        if self._read_escape:
            self._read_escape = False
            self.is_escaped = True
        else:
            self.is_escaped = False
        if not self.is_escaped and c == "%":
            self.in_comment = True
        if c == "\n":
            self.in_comment = False
        if not self.in_comment and not self.is_escaped and c == "\\":
            assert self._read_escape is False
            self._read_escape = True
        else:
            self._read_escape = False
        return c

    def advance(self, n: int = 1) -> None:
        """
        Advance the cursor by n characters.
        :param n: Number of characters to advance.
        :return: None
        """
        for _ in range(n):
            try:
                self.next()
            except _ParserError:
                break

    def peek(self, pure=False) -> typing.Optional[str]:
        """
        Return the current character. Do not move the cursor.
        :param pure: Only return characters that are not in a comment or escaped.
        :return: Next character.
        """
        if pure and (self._read_escape or self.in_comment):
            return None
        if not self.has_next():
            return None
        c = self._text[self._pos]
        return c

    def has_next(self) -> bool:
        """
        Returns whether there is a next character.
        :return: True if there is a next character.
        """
        return self._pos < len(self._text)

    def pos(self) -> int:
        """
        Current position of the cursor.
        :return: Position of the cursor.
        """
        return self._pos

    def skip_whitespace_and_comments(self) -> None:
        """
        Skips over all whitespace characters and comments.
        :return: None
        """
        while self.in_comment or (self.peek().isspace() and not self.is_escaped):
            self.next()

    def __iter__(self):
        """
        Iterate over all characters. You can use the other methods during the
        iteration, allowing you for example to skip all whitespaces and comments
        in the loop.
        :return: All characters.
        """
        while self.has_next():
            c = self.next()
            yield c


class CommandMatch:
    """
    A match of the CommandFinder.
    """

    def __init__(
        self,
        command: str,
        start: int,
        end: int,
        parameters: typing.List[typing.Tuple[int, int]],
        opt_parameters: typing.List[typing.Optional[typing.Tuple[int, int]]],
    ):
        self.command = command  # command name
        self.start = start  # position of the start of the command
        self.end = end  # position after the last parameter of the command
        self.parameters = parameters  # list of the mandatory parameters
        self.opt_parameters = opt_parameters  # list of the optional parameters

    def __repr__(self):
        return (
            f"{self.start}:{self.end} \\{self.command}"
            + "["
            + "".join(f"{p[0]}:{p[1]}" if p else "None" for p in self.opt_parameters)
            + "]"
            + "".join("{" + str(p[0]) + ":" + str(p[1]) + "}" for p in self.parameters)
        )

    def __eq__(self, other):
        if other is None:
            return False
        return (
            self.command,
            self.start,
            self.end,
            self.parameters,
            self.opt_parameters,
        ) == (
            other.command,
            other.start,
            other.end,
            other.parameters,
            other.opt_parameters,
        )


class CommandFinder:
    """
    Finds occurrences of latex commands in a string and provides you with
    the position and nicely parsed parameters.
    """

    def __init__(self, strict=False):
        self._strict = strict
        self._commands = {}

    def add_command(self, name, num_params=1, num_opt=0):
        """
        :param name:
        :param num_params:
        :param num_opt:
        :return:
        """
        self._commands[name] = (num_params, num_opt)
        return self

    def _read_parameters(self, stream, name: str):
        if name not in self._commands:
            return [], []
        n, n_opt = self._commands[name]
        opt_params = [
            self._read_parameter(stream, "[", "]", False) for _ in range(n_opt)
        ]
        params = [self._read_parameter(stream, "{", "}") for _ in range(n)]
        return opt_params, params

    def _read_command_name(self, stream):
        stream.skip_whitespace_and_comments()
        if stream.peek(True) != "\\":
            msg = f"No command. Next character is '{stream.peek()}'."
            raise _ParserError(msg, stream.pos())
        command = ""
        stream.next()  # skip '\'
        while stream.has_next() and (stream.peek().isalpha() or stream.peek() == "*"):
            command += stream.next()
        return command

    def _read_parameter(
        self, stream: LatexStream, begin: str, end: str, mandatory=True
    ):
        stream.skip_whitespace_and_comments()
        if not mandatory and stream.peek(True) != begin:
            # No begin-symbol ({[) -> no parameter if not mandatory
            return None
        if stream.peek(True) == begin:  # properly encapsulated parameter.
            depth = 1
            stream.advance()
            start = stream.pos()  # after {[
            while depth > 0:
                if stream.peek(True) == begin:
                    depth += 1
                if stream.peek(True) == end:
                    depth -= 1
                stream.advance()
            return (start, stream.pos() - 1)  # at }]
        # parameter without begin/end-symbols ([],{})
        if self._strict:
            context = stream._text[stream.pos() - 10 : stream.pos() + 10]
            msg = f"Parameters without brackets ('{context}')."
            raise _ParserError(msg, stream.pos())
        start = stream.pos()
        if stream.peek() == "\\":
            context = stream._text[stream.pos() - 10 : stream.pos() + 10]
            logging.getLogger("flachtex").warning(
                f"Ambiguous parameters due to missing brackets ('{context}')."
                f" This can lead to corruptions."
            )
            command_name = self._read_command_name(stream)
            self._read_parameters(stream, command_name)
            return (start, stream.pos())
        stream.advance()
        return (start, stream.pos())

    def _read_new_command_parameters(self, stream: LatexStream):
        command_name = self._read_parameter(stream, "{", "}", mandatory=True)
        opt_params = [self._read_parameter(stream, "[", "]", mandatory=False)]
        definition = self._read_parameter(stream, "{", "}", mandatory=True)
        return opt_params, [command_name, definition]

    def find(self, text: str, begin: int = 0) -> typing.Optional[CommandMatch]:
        """
        Find first occurrence of a command in the text.
        :param text: The text to be searched.
        :param begin: The point to start in the text.
        :return:
        """
        stream = LatexStream(text, begin)
        while stream.has_next():
            try:
                if stream.peek(True) == "\\":
                    begin = stream.pos()
                    command = self._read_command_name(stream)
                    if command in (
                        "newcommand",
                        "renewcommand",
                        "newcommand*",
                        "renewcommand*",
                    ):
                        if command in self._commands:
                            opt_params, params = self._read_new_command_parameters(
                                stream
                            )
                            end = stream.pos()
                            return CommandMatch(command, begin, end, params, opt_params)
                        #  In the \\newcommand definition, the commands are not actually
                        # applied, so we want to skip them.
                        self._read_parameter(stream, "{", "}")  # skip definition name
                    elif command in self._commands:
                        opt_params, params = self._read_parameters(stream, command)
                        end = stream.pos()
                        return CommandMatch(command, begin, end, params, opt_params)

                else:
                    stream.next()
            except _ParserError as pe:
                _logger.error(str(pe))
                stream.advance()
        return None

    def find_all(self, text: str):
        begin = 0
        result = self.find(text, begin)
        while result:
            yield result
            begin = result.end
            result = self.find(text, begin)



"""
Some parts of the LaTeX-document may need to be substituted.
The rules defined in this file allow such a handling.
"""

import abc
import logging
import typing

from flachtex.command_finder import CommandFinder
from flachtex.traceable_string import TraceableString
from flachtex.utils import Range


class Substitution(Range):
    def __init__(
        self, start: int, end: int, replacement_text: typing.Optional[TraceableString]
    ):
        super().__init__(start, end)
        assert not replacement_text or isinstance(replacement_text, TraceableString)
        self.replacement_text = replacement_text

    def __repr__(self):
        return f"REPLACEMENT[{self.start}:{self.end}]{{{self.replacement_text}}}"


class SubstitutionRule(abc.ABC):
    @abc.abstractmethod
    def find_all(self, content: TraceableString) -> typing.Iterable[Substitution]:
        pass


class ChangesRule(SubstitutionRule):
    """
    The changes-package (\\usepackage{changes}) allows to highlight changes in the
    current revision (especially for writing journal papers). This rule allows
    to remove all its usage and is actually more robust than the original script
    supplied with the package.
    """

    def __init__(self, prefix=False):
        self._added = "chadded" if prefix else "added"
        self._deleted = "chdeleted" if prefix else "deleted"
        self._replaced = "chreplaced" if prefix else "replaced"
        self._highlight = "chhighlight" if prefix else "highlight"
        self._comment = "chcomment" if prefix else "comment"

    def find_all(self, content: TraceableString) -> typing.Iterable[Substitution]:
        assert isinstance(content, TraceableString)
        cf = CommandFinder()
        cf.add_command(self._added, 1, 1)
        cf.add_command(self._deleted, 1, 1)
        cf.add_command(self._replaced, 2, 1)
        cf.add_command(self._highlight, 1, 1)
        cf.add_command(self._comment, 1, 1)
        for match in cf.find_all(str(content)):
            if match.command in (self._added, self._replaced, self._highlight):
                yield Substitution(
                    match.start,
                    match.end,
                    content[match.parameters[0][0] : match.parameters[0][1]],
                )
            elif match.command in (self._deleted, self._comment):
                yield Substitution(match.start, match.end, None)


def _sort_replacements(
    replacements: typing.List[Substitution],
) -> typing.Iterable[Substitution]:
    replacements.sort()
    if len(replacements) <= 1:
        return replacements
    replacements_ = []
    for i, e in enumerate(replacements):
        if e.intersects(replacements[(i + 1) % len(replacements)]):
            continue
        replacements_ += [e]
    return replacements_


def _find_substitutions(
    content: TraceableString, replacement_rules: typing.List[SubstitutionRule]
) -> typing.List[Substitution]:
    replacements = []
    for rule in replacement_rules:
        replacements += list(rule.find_all(content))
    return replacements


def apply_substitution_rules(
    content: TraceableString,
    replacement_rules: typing.List[SubstitutionRule],
):
    replacements = _find_substitutions(content, replacement_rules)
    max_itererations = 10
    while replacements and max_itererations:
        replacements = _sort_replacements(replacements)
        offset = 0
        for replacement in replacements:
            if replacement.replacement_text:
                content = (
                    content[: replacement.start + offset]
                    + replacement.replacement_text
                    + content[replacement.end + offset :]
                )
                offset -= len(replacement) - len(replacement.replacement_text)
            else:
                content = (
                    content[: replacement.start + offset]
                    + content[replacement.end + offset :]
                )
                offset -= len(replacement)
        max_itererations -= 1
        replacements = _find_substitutions(content, replacement_rules)
    if max_itererations == 0:
        logging.getLogger("flachtex").warning(
            "Exceeded maximal replacement iterations."
        )
    return content



import os.path
import typing
from pathlib import Path


class FileSystem:
    """
    Wraps the file system access such that it could be replaced with a simple dict
    to ease testing.
    """

    def __contains__(self, item) -> bool:
        return os.path.exists(item) and os.path.isfile(item)

    def __getitem__(self, item) -> str:
        if item not in self:
            msg = f"Could not find {item}."
            raise KeyError(msg)
        with Path(item).open(errors="ignore") as f:
            try:
                return str(f.read())
            except Exception as e:
                return f"\n%ERROR (flachtex): Could not read '{item}': '{e}'\n"


class FileFinder:
    def __init__(self, project_root=".", file_system=None):
        """
        :param project_root: The root of the project (relative to cwd)
        :param file_system: A module to open files, can be replaced with a simple dict
            for simple testing.
        """
        if not file_system:
            file_system = FileSystem()
        self.file_system = file_system
        self._PATH = [project_root]
        self._project_root = project_root

    def set_root(self, project_root: str):
        self._PATH = [project_root]

    def find_best_matching_path(self, path: str, origin: str) -> str:
        """
        Returns the best path relative to the current working directory that resolves the
        wanted path, which is not necessarily relative to the current working directory.
        :param path: Path to be resolved.
        :param origin: The path to the file (relative to cwd) that tries to open above's
            path.
        :return:
        """
        for p in self.get_checked_paths(path, origin):
            if p in self.file_system:
                return p
        msg = f"Not matching file found. Tried: {', '.join(self.get_checked_paths(path, origin))}"
        raise KeyError(msg)

    def _normalize(self, path: str):
        return os.path.normpath(path)

    def get_checked_paths(self, path: str, origin: str) -> typing.Iterable[str]:
        """
        Returns all paths that will be tried to find the file.
        :param path: Path to the file. Not necessarily regarding the current working
                        directory.
        :param origin: Path of the file that tries to access above's path
        :return: The best matching path to the file relative to the current working
                    directory, i.e., can be opened directly.
        """
        # if it is an absolute path, try this one first
        if os.path.isabs(path):
            yield os.path.normpath(path)
            yield os.path.normpath(path) + ".tex"
        # then try to go relative from the origin file
        d = os.path.dirname(origin)
        yield self._normalize(os.path.join(d, path))
        yield self._normalize(os.path.join(d, path)) + ".tex"
        # then try to use the include directories
        for include in self._PATH:
            yield self._normalize(os.path.join(include, path))
            yield self._normalize(os.path.join(include, path)) + ".tex"
        # finally, in a last attempt, go upwards from the origin file
        while d != self._project_root:  # stop if the root directory has been reached
            yield self._normalize(os.path.join(d, path))
            yield self._normalize(os.path.join(d, path)) + ".tex"
            d = os.path.dirname(d)  # go one directory above

    def read(self, path) -> str:
        """
        Just returns the content of the path as a single string.
        :param path: Path to the file
        :return: File content
        """
        return self.file_system[self._normalize(path)]



import pytest

from flachtex import TraceableString


def test_get_origin():
    ts = TraceableString("content", "origin", 0)
    assert ts.get_origin(2) == ("origin", 2)


def test_add():
    ts1 = TraceableString("left", "A", 0)
    ts2 = TraceableString("right", "B", 0)
    ts = ts1 + ts2
    assert str(ts) == "leftright"


def test_json():
    ts1 = TraceableString("left", "A", 0)
    ts2 = TraceableString("right", "B", 0)
    ts = ts1 + ts2
    ts_ = TraceableString.from_json(ts.to_json())
    assert ts == ts_


def test_invalid_json():
    ts1 = TraceableString("left", "A", 0)
    ts2 = TraceableString("right", "B", 0)
    ts = ts1 + ts2
    data = ts.to_json()
    data["origins"].append([])  # Introduce invalid data
    with pytest.raises(ValueError):
        TraceableString.from_json(data)



# flake8: noqa F401
from .filefinder import FileFinder
from .traceable_string import TraceableString
from .comments import remove_comments
from .command_finder import CommandFinder
from .preprocessor import Preprocessor

__all__ = [
    "FileFinder",
    "TraceableString",
    "remove_comments",
    "CommandFinder",
    "Preprocessor",
]



from flachtex import FileFinder, Preprocessor
from flachtex.rules import SubimportChangesRule


def flatten(document, root="main.tex"):
    preprocessor = Preprocessor("/")
    file_finder = FileFinder("/", document)
    preprocessor.file_finder = file_finder
    preprocessor.subimport_rules.append(SubimportChangesRule())
    return str(preprocessor.expand_file(root))


def test_no_import():
    document = {
        "main.tex": "line 0\nline 1\n\\line 2\nline 3\nline 4\n",
        "sub.tex": "line 2\nline 3\n",
    }
    assert flatten(document) == "line 0\nline 1\n\\line 2\nline 3\nline 4\n"


def test_single_import():
    document = {
        "main.tex": "line 0\nline 1\n\\input{sub.tex}\nline 4\n",
        "sub.tex": "line 2\nline 3",
    }
    assert flatten(document) == "line 0\nline 1\nline 2\nline 3\nline 4\n"


def test_two_imports():
    document = {
        "main.tex": "line 0\nline 1\n\\input{sub.tex}\n\\input{sub2.tex}\nline 4\n",
        "sub.tex": "line 2\n",
        "sub2.tex": "line 3\n",
    }
    assert flatten(document) == "line 0\nline 1\nline 2\n\nline 3\n\nline 4\n"


def test_single_import_with_linebreak():
    document = {
        "main.tex": "line 0\nline 1\n\\input{\nsub.tex\n}\nline 4\n",
        "sub.tex": "line 2\nline 3",
    }
    assert flatten(document) == "line 0\nline 1\nline 2\nline 3\nline 4\n"


def test_import_from_subfolder():
    document = {
        "main.tex": "line 0\nline 1\n\\input{dir/sub.tex}\nline 4\n",
        "dir/sub.tex": "line 2\nline 3",
    }
    assert flatten(document) == "line 0\nline 1\nline 2\nline 3\nline 4\n"


def test_relative_import_with_subfolder():
    document = {
        "main.tex": "line 0\nline 1\n\\input{dir/sub.tex}\nline 4\n",
        "dir/sub.tex": "line 2\n\\include{sub2.tex}\n",
        "dir/sub2.tex": "line 3\n",
    }
    assert flatten(document) == "line 0\nline 1\nline 2\nline 3\n\n\nline 4\n"


def test_subimport():
    document = {
        "main.tex": "line 0\nline 1\n\\subimport{dir}{sub}\nline 4\n",
        "dir/sub.tex": "line 2\nline 3",
    }
    assert flatten(document) == "line 0\nline 1\nline 2\nline 3\nline 4\n"


def test_multiple_subimports():
    document = {
        "main.tex": "line 0\nline 1\n\\subimport{dir}{sub}\n\\subimport{dir}{sub2}\nline 4\n",
        "dir/sub.tex": "line 2",
        "dir/sub2.tex": "line 3",
    }
    assert flatten(document) == "line 0\nline 1\nline 2\nline 3\nline 4\n"


def test_subimport_with_star():
    document = {
        "main.tex": "line 0\nline 1\n\\subimport*{dir}{sub}\nline 4\n",
        "dir/sub.tex": "line 2\nline 3",
    }
    assert flatten(document) == "line 0\nline 1\nline 2\nline 3\nline 4\n"


def test_multiple_subimports_with_star():
    document = {
        "main.tex": "line 0\nline 1\n\\subimport*{dir}{sub}\n\\subimport*{dir}{sub2}\nline 4\n",
        "dir/sub.tex": "line 2",
        "dir/sub2.tex": "line 3",
    }
    assert flatten(document) == "line 0\nline 1\nline 2\nline 3\nline 4\n"


def test_skip_functionality():
    document = {
        "main.tex": "line 0\n%%FLACHTEX-SKIP-START\nline 3\n%%FLACHTEX-SKIP-STOP\nline 5\n"
    }
    assert flatten(document) == "line 0\n\nline 5\n"


def test_multiple_skip_blocks():
    document = {
        "main.tex": "line 0\n%%FLACHTEX-SKIP-START\nline 3\n%%FLACHTEX-SKIP-STOP\n%%FLACHTEX-SKIP-START\nbla\n%%FLACHTEX-SKIP-STOP\nline 5\n"
    }
    assert flatten(document) == "line 0\n\n\nline 5\n"


def test_complex_subimports():
    document = {
        "main.tex": "line 0\nline 1\n\\input{sub.tex}\nline 4\n\\subimport{subimport}{subimport.tex}\nline 10\n",
        "sub.tex": "line 2\nline 3",
        "subimport/subimport.tex": "line 5\n\\includegraphics{1.pdf}\n\\includegraphics{/2.pdf}\n\\includegraphics{./3.pdf}\nline 6\n\\include{./subimportb.tex}\nline 9\n",
        "subimport/subimportb.tex": "line 7\n\\includegraphics{4.pdf}\nline 8\n",
    }
    assert flatten(document) == (
        "line 0\nline 1\nline 2\nline 3\nline 4\n"
        "line 5\n\\includegraphics{./subimport/1.pdf}\n\\includegraphics{./subimport/2.pdf}\n\\includegraphics{./subimport/3.pdf}\nline 6\n"
        "line 7\n\\includegraphics{./subimport/4.pdf}\nline 8\n\nline 9\n"
        "\nline 10\n"
    )



from flachtex import FileFinder
from flachtex.preprocessor import Preprocessor


def flatten(document, root="main.tex"):
    preprocessor = Preprocessor("/")
    file_finder = FileFinder("/", document)
    preprocessor.file_finder = file_finder
    return preprocessor.expand_file(root), preprocessor.structure


def test_no_import():
    test_document = {
        "main.tex": "line 0\nline 1\nline 2\nline 3\nline 4\n",
    }
    flat, sources = flatten(test_document)
    for path, source in test_document.items():
        assert source == sources[path]["content"]
    assert flat.get_origin_of_line(0, 0) == ("main.tex", 0)
    assert flat.get_origin(0) == ("main.tex", 0)
    assert flat.get_origin_of_line(0, 4) == ("main.tex", 4)
    assert flat.get_origin(4) == ("main.tex", 4)
    assert flat.get_origin_of_line(1, 0) == ("main.tex", 7)
    assert flat.get_origin_of_line(2, 4) == ("main.tex", 18)
    assert flat.get_origin(17) == ("main.tex", 17)


def test_single_import():
    test_document = {
        "main.tex": "line 0\nline 1\n\\input{sub.tex}\nline 4\n",
        "sub.tex": "line 2\nline 3",
    }
    flat, sources = flatten(test_document)
    for path, source in test_document.items():
        assert source == sources[path]["content"]
    assert flat.get_origin_of_line(0, 0) == ("main.tex", 0)
    assert flat.get_origin(0) == ("main.tex", 0)
    assert flat.get_origin_of_line(0, 4) == ("main.tex", 4)
    assert flat.get_origin(4) == ("main.tex", 4)
    assert flat.get_origin_of_line(2, 4) == ("sub.tex", 4)
    assert flat.get_origin(17) == ("sub.tex", 3)



from flachtex import CommandFinder
from flachtex.command_finder import CommandMatch


def test_empty():
    cf = CommandFinder()
    cf.add_command("todo", 1)
    text = ""
    assert cf.find(text) is None


def test_no_match():
    cf = CommandFinder()
    cf.add_command("todo", 1)
    text = "This is a simple string\n bla"
    assert cf.find(text) is None


def test_no_match2():
    cf = CommandFinder()
    cf.add_command("todo", 1)
    text = "This is a \\todonot{bla} simple string\n bla"
    assert cf.find(text) is None


def test_valid_command():
    cf = CommandFinder()
    cf.add_command("todo", 1)
    text = "This is a \\todo{bla} simple string\n bla"
    m = cf.find(text)
    assert m == CommandMatch("todo", 10, 20, [(16, 19)], [])
    assert text[m.start : m.end] == "\\todo{bla}"


def test_optional_param():
    cf = CommandFinder()
    cf.add_command("todo", 1, 1)
    text = "This is a \\todo{bla} simple string\n bla"
    assert cf.find(text) == CommandMatch("todo", 10, 20, [(16, 19)], [None])


def test_hash_param():
    cf = CommandFinder()
    cf.add_command("todo", 1, 1)
    text = "This is a \\todo{bla#1} simple string\n bla"
    assert cf.find(text) == CommandMatch("todo", 10, 22, [(16, 21)], [None])


def test_newline_in_param():
    cf = CommandFinder()
    cf.add_command("todo", 1)
    text = "This is a \\todo{bla\nbla} simple string\n bla"
    assert cf.find(text) == CommandMatch("todo", 10, 24, [(16, 23)], [])


def test_comment_in_param():
    cf = CommandFinder()
    cf.add_command("todo", 1)
    text = "This is a \\todo{bla%}\nbla} simple string\n bla"
    assert cf.find(text) == CommandMatch("todo", 10, 26, [(16, 25)], [])


def test_strict_mode_bad_command():
    cf = CommandFinder(strict=True)
    cf.add_command("renewcommand", 2)
    text = "This is a simple string\n bla \\renewcommand\\thesubfigure{(\\alph{subfigure})} asdas"
    assert cf.find(text) is None



from flachtex import TraceableString
from flachtex.command_substitution import (
    NewCommandDefinition,
    NewCommandSubstitution,
    find_new_commands,
)
from flachtex.rules import apply_substitution_rules

doc1 = TraceableString(
    r"""
\newcommand{\MSClength}{MSC-MS\xspace}
\newcommand{\MSCtotEnergy}{MSC-TE\xspace}
\newcommand{\MSClocEnergy}{MSC-BE\xspace}

\newcommand{\MIPone}{MIP-1\xspace}
\newcommand{\MIPtwo}{MIP-2\xspace}
\newcommand*{\MIPthree}{MIP-3\xspace}
\newcommand{\CPone}{CP-1\xspace}
\newcommand{\CPtwo}{CP-2\xspace}

\newcommand{\cone}{cone\xspace}\newcommand{\cupdot}{\mathbin{\mathaccent\cdot\cup}}
\newcommand*{\mincover}{$\Lambda$-cover\xspace}

\MIPone
""",
    "main.tex",
)


def test_empty():
    cmds = list(find_new_commands(TraceableString("", "Main.tex")))
    assert cmds == []


def test_doc1():
    cmds = list(find_new_commands(doc1))
    assert len(cmds) == 11
    sub = NewCommandSubstitution()
    for cmd in cmds:
        sub.new_command(NewCommandDefinition(cmd.name, cmd.num_parameters, cmd.command))
    replacements = list(sub.find_all(doc1))
    assert len(replacements) == 1


def test_substitution_with_zero_parameters():
    sub = NewCommandSubstitution()
    sub.new_command(
        NewCommandDefinition(
            TraceableString("test", None), 0, TraceableString("TEST\\xspace", None)
        )
    )
    s = apply_substitution_rules(
        TraceableString("Bla \\test asd \\test{}.", None), [sub]
    )
    assert str(s) == "Bla TEST\\xspace asd TEST\\xspace{}."


def test_substitution_with_non_space_replacement():
    sub = NewCommandSubstitution()
    sub.new_command(
        NewCommandDefinition(
            TraceableString("test", None), 0, TraceableString("TEST", None)
        )
    )
    s = apply_substitution_rules(
        TraceableString("Bla \\test asd \\test{}.", None), [sub]
    )
    assert str(s) == "Bla TEST{}asd TEST{}."


def test_complex_commands():
    doc2 = TraceableString(
        r"""
\newcommand{\test}[2]{#2-#1}
\newcommand{\testa}[2]{#2-#1.}
\newcommand{\testb}[2]{#2#1.}
\begin{document}
\test{a}{b}.\\
\testa{a}{b}.\\
\testb{a}{b}.
\testb{aa}{bb}.
\end{document}
""",
        "main.tex",
    )
    cmds = list(find_new_commands(doc2))
    assert len(cmds) == 3
    sub = NewCommandSubstitution()
    for cmd in cmds:
        sub.new_command(NewCommandDefinition(cmd.name, cmd.num_parameters, cmd.command))
    replacements = list(sub.find_all(doc2))
    assert len(replacements) == 4
    s = apply_substitution_rules(doc2, [sub])
    assert (
        str(s).strip()
        == "\\newcommand{\\test}[2]{#2-#1}\n\\newcommand{\\testa}[2]{#2-#1.}\n\\newcommand{\\testb}[2]{#2#1.}\n\\begin{document}\nb-a.\\\\\nb-a..\\\\\nba..\nbbaa..\n\\end{document}".strip()
    )



"""
Removing comments can only be done at the end as comments can be used for
directives for flachtex. Removing them already in the preprocessor would potentially
lead to unwanted effects.
If you don't want comments, simply apply the function below to the output.
"""

import re

from .traceable_string import TraceableString


def remove_comments(content: TraceableString) -> TraceableString:
    """
    Remove the comments out of the content.
    :param content:
    :return:
    """
    regex = re.compile(r"^.*?(?<!\\)(?P<comment>%..*\n)", re.MULTILINE)
    comments = []
    for match in regex.finditer(str(content)):
        comments.append((match.start("comment"), match.end("comment")))
    comments.sort()
    offset = 0
    for comment in comments:
        content = content[: comment[0] + offset] + content[comment[1] + offset :]
        offset -= comment[1] - comment[0]
    return content



class CycleException(Exception):
    def __init__(self, path, origin):
        self.path = path
        self.origin = origin

    def __str__(self):
        if self.origin:
            return f"CycleException importing {self.path} ({self.origin})."
        return f"CycleException importing {self.path}."


class CyclePrevention:
    def __init__(self):
        self._checked_paths = []

    def push(self, path, context=None):
        if path in self._checked_paths:
            raise CycleException(path, context)
        self._checked_paths.append(path)

    def pop(self):
        self._checked_paths.pop()



import abc
import os
import re
import typing

from flachtex.traceable_string import TraceableString
from flachtex.utils import Range


class Import(Range):
    def __init__(
        self, start: int, end: int, path: str, is_subimport: bool, subimport_path: str
    ):
        super().__init__(start, end)
        self.path = path
        self.is_subimport = is_subimport if not None else False
        self.subimport_path = subimport_path


class ImportRule(abc.ABC):
    @abc.abstractmethod
    def find_all(self, content: str) -> typing.Iterable[Import]:
        pass


class RegexImportRule(ImportRule):
    def __init__(self, regex: str):
        self.regex = re.compile(regex, re.MULTILINE | re.DOTALL)

    @abc.abstractmethod
    def determine_include(self, match: re.Match) -> Import:
        pass

    def find_all(self, content: str) -> typing.Iterable[Import]:
        for match in self.regex.finditer(content):
            yield self.determine_include(match)


class NativeImportRule(RegexImportRule):
    """
    Detects includes of the form `\\input{/path/file.tex}` and `\\include{/path/file.tex}`
    """

    def __init__(self):
        expr = r"^(([^%\n])|(\\%))*(?P<command>((\\input)|(\\include))\{(?P<path>[^}]*?)\})"
        super().__init__(expr)

    def determine_include(self, match: re.Match):
        import_path = match.group("path").strip()
        return Import(
            match.start("command"), match.end("command"), import_path, False, None
        )


class SubimportRule(RegexImportRule):
    """
    Detects imports by the subimport package.
    These can have the form `\\subimport{path}{file}` or  `\\subimport*{path}{file}`.
    """

    expr = r"^(([^%])|(\\%))*?(?P<command>\\subimport\*?\{(?P<dir>[^}]*)\}\{(?P<file>[^}]*)\})"

    def __init__(self):
        super().__init__(self.expr)

    def determine_include(self, match: re.Match):
        # This function implements the functionality for the subimports library.
        # The import is separated into two parts
        subimport_path = match.group("dir").strip()
        import_path = os.path.join(
            match.group("dir").strip(), match.group("file").strip()
        )
        return Import(
            match.start("command"),
            match.end("command"),
            import_path,
            True,
            subimport_path,
        )


# %%FLACHTEX-EXPLICIT-IMPORT[path/file.tex]
class ExplicitImportRule(RegexImportRule):
    def __init__(self):
        super().__init__(
            r"^\s*(?P<command>%%FLACHTEX-EXPLICIT-IMPORT\[(?P<path>[^}]*)\])"
        )

    def determine_include(self, match: re.Match):
        # We are using the group feature of regex to extract the path (<path>)
        # as well as the part to be replaced (<command>)
        import_path = match.group("path").strip()
        return Import(
            match.start("command"), match.end("command"), import_path, False, None
        )


def _sort_imports(imports: typing.List[Import]) -> typing.List[Import]:
    imports.sort()
    for i, e in enumerate(imports[:-1]):
        if e.intersects(imports[i + 1]):
            msg = "Intersecting imports."
            raise ValueError(msg)
    return imports


def find_imports(
    content: TraceableString, include_rules: typing.Iterable[ImportRule]
) -> typing.List[Import]:
    content = str(content)
    imports = []
    for rule in include_rules:
        imports += list(rule.find_all(content))
    imports = _sort_imports(imports)
    return imports



"""
LaTeX allows you to define your own commands using ``\\newcommand``.
This can be a useful feature for many scenarios, but it complicates the source.
Thus, this file implements some logic to substitute these usages by their definition.
"""

import logging
import re
import typing

from flachtex.command_finder import CommandFinder
from flachtex.rules import Substitution, SubstitutionRule
from flachtex.traceable_string import TraceableString

_logger = logging.getLogger(__file__)


class NewCommandDefinition:
    """
    For defining a LaTeX-command definition with ``\\newcommand``. Note that optional
    parameters are not supported right now.
    """

    def __init__(
        self, name: TraceableString, num_parameters: int, command: TraceableString
    ):
        """
        :param name: The name of the command.
        :param num_parameters: The number of (mandatory parameters).
        :param command: The actual command definition.
        """
        if not isinstance(command, TraceableString):
            _logger.warning("Command is not traceable!")
            command = TraceableString(str(command), None)
        self.name = name
        self.num_parameters = num_parameters
        self.command = command

    def __repr__(self):
        return f"\\newcommand{{{self.name}}}[{self.num_parameters}]{{{self.command}}}"


def find_new_commands(
    latex_document: TraceableString,
) -> typing.Iterable[NewCommandDefinition]:
    """
    Find all commands defined by ``\\newcommand``. Not compatible with optional commands
    right now.
    :param latex_document: The LaTeX-document to be scanned.
    :return: Iterator on all command definitions.
    """
    cf = CommandFinder(strict=True)
    cf.add_command("newcommand", 2, 1)
    cf.add_command("newcommand*", 2, 1)
    cf.add_command("renewcommand", 2, 1)
    cf.add_command("renewcommand*", 2, 1)
    for match in cf.find_all(str(latex_document)):
        command_name = latex_document[match.parameters[0][0] : match.parameters[0][1]]
        command = latex_document[match.parameters[1][0] : match.parameters[1][1]]
        if match.opt_parameters[0]:
            opt_par_range = match.opt_parameters[0]
            num_parameters = int(
                str(latex_document[opt_par_range[0] : opt_par_range[1]])
            )
        else:
            num_parameters = 0
        yield NewCommandDefinition(command_name, num_parameters, command)


class NewCommandSubstitution(SubstitutionRule):
    """
    Substitute commands defined, e.g., by \newcommand.
    Currently, default parameters are not supported.
    """

    def __init__(self, space_substitution: bool = True):
        """
        :param space_substitution: Try to simulate the missing space of parameterless commands.
        """
        self._commands = {}
        self._command_finder = CommandFinder()
        self._space_sub = space_substitution

    def new_command(self, definition: NewCommandDefinition) -> None:
        """
        Add a new command definition that will be replaced.
        :param definition: The definition of the command
        :return: None
        """
        name = str(definition.name).strip()
        if name[0] == "\\":
            name = name[1:]
        if name in self._commands:
            _logger.warning(
                f"Multiple definitions of command '{name}'. Substitution may be buggy."
            )
        self._commands[name] = definition
        _logger.info(f"Detected {definition}")
        self._command_finder.add_command(name, definition.num_parameters)

    def _get_substitution(
        self, command: TraceableString, parameters: typing.List[TraceableString]
    ) -> TraceableString:
        for i, p in enumerate(parameters):
            i += 1
            offset = 0
            for match in re.finditer(
                f"(?P<arg>#{i})([^0-9])", str(command) + " ", re.MULTILINE
            ):
                l = (match.end("arg") - match.start("arg")) - 1
                command = (
                    command[: match.start("arg") + offset]
                    + p
                    + command[(match.end("arg")) + offset :]
                )
                offset += len(p) - l
        return command

    def find_all(self, content: TraceableString) -> typing.Iterable[Substitution]:
        for match in self._command_finder.find_all(str(content)):
            definition = self._commands[str(match.command)]
            parameters = [content[p[0] : p[1]] for p in match.parameters]
            sub = self._get_substitution(definition.command, parameters)
            end = match.end
            if (
                self._space_sub
                and definition.num_parameters == 0
                and not str(sub).strip().endswith("\\xspace")
            ):
                # The usage of a command like "\\cmd bla" is actually equivalent to
                # "\\cmd{}bla". This function tries to simulate this.
                while content[end] == " ":
                    end += 1
                if end != match.end:
                    sub += TraceableString("{}", None)  # add non-space separator
            yield Substitution(match.start, end, sub)



import typing
import unittest


class Range:
    """
    A simple range (for within a text)
    """

    def __init__(self, start: int, end: int):
        self.start = start
        self.end = end

    def intersects(self, other):
        # one begin lies within the other
        if self.start <= other.start < self.end:
            return True
        if other.start <= self.start < self.end and not other.end <= self.start:
            return True
        return False

    def __le__(self, other):
        return self.start <= other.start

    def __lt__(self, other):
        return self.start < other.start

    def __len__(self):
        return self.end - self.start

    def __repr__(self):
        return f"[{self.start}:{self.end}]"


def compute_row_index(content: str) -> typing.List[int]:
    index = [0]
    i = content.find("\n")
    while i >= 0:
        index.append(i + 1)
        i = content.find("\n", i + 1)
    return index


class TestRowIndex(unittest.TestCase):
    def test_1(self):
        text = "0\n1\n2\n3\n4\n"
        index = compute_row_index(text)
        for i in text.split("\n"):
            b = text.find(i)
            if i:
                assert index[int(i)] == b
        print(index)

    def test_2(self):
        text = "\n1\n2\n3\n4\n"
        index = compute_row_index(text)
        for i in text.split("\n"):
            b = text.find(i)
            if i:
                assert index[int(i)] == b
        print(index)



import typing

from .utils import compute_row_index


class OriginOfRange:
    def __init__(self, begin: int, end: int, origin, offset: int = 0):
        self.origin = origin
        self.begin = begin
        self.end = end
        self.offset = offset

    def __len__(self):
        return self.end - self.begin

    def cut_front(self, n: int):
        """
        All indices <n shall be cut off.
        :param n: New first position
        :return: Modified OriginOfRange corresponding to new beginning. If this origin is
        no longer contained, return None.
        """
        if self.end <= n:
            return None
        if self.begin <= n < self.end:
            offset_ = self.offset + (n - self.begin)
        else:
            offset_ = self.offset
        return OriginOfRange(max(0, self.begin - n), self.end - n, self.origin, offset_)

    def cut_back(self, n: int):
        """
        All indices >=n shall be cut off.
        :param n: New end index
        :return: Modified OriginOFRange corresponding to shortened string. If this origin is
        no longer contained, return None.
        """
        if self.begin >= n:
            return None
        return OriginOfRange(self.begin, min(self.end, n), self.origin, self.offset)

    def slice(self, begin, end):
        # cutting of the back does not change the indices and should be done first.
        shortened = self.cut_back(end)
        if shortened:
            return shortened.cut_front(begin)
        return shortened

    def __lt__(self, other):
        return self.begin < other.start

    def get_offset(self, i):
        if i not in self:
            msg = "Index is not within this origin range."
            raise ValueError(msg)
        return self.offset + (i - self.begin)

    def __contains__(self, item):
        return self.begin <= item < self.end

    def move(self, n):
        return OriginOfRange(self.begin + n, self.end + n, self.origin, self.offset)

    def __repr__(self):
        return f"{self.origin}[{self.begin}:{self.end}:+{self.offset}]"

    def __eq__(self, other):
        if not isinstance(other, OriginOfRange):
            return False
        return (
            other.begin == self.begin
            and self.end == other.end
            and self.origin == other.origin
            and self.offset == other.offset
        )


class TraceableString:
    def __init__(self, content: str, origin: typing.Any, offset: int = 0):
        self.content = content
        self.origins = [OriginOfRange(0, len(content), origin, offset)]
        self._line_index = None

    def __len__(self):
        return len(self.content)

    def __str__(self):
        return self.content

    def _normalize_slice(self, s: slice):
        start = s.start if s.start is not None else 0
        start = start if start >= 0 else len(self) - start
        stop = s.stop if s.stop is not None else len(self)
        stop = stop if stop >= 0 else len(self) - stop
        if stop > len(self):
            raise IndexError()
        if s.step is not None and s.step != 1:
            msg = "Slicing with steps is not supported."
            raise NotImplementedError(msg)
        return start, stop

    def __getitem__(self, item):
        if isinstance(item, slice):
            content = self.content[item]
            start, stop = self._normalize_slice(item)
            origins = (o.slice(start, stop) for o in self.origins)
            origins = [o for o in origins if o is not None]
            ts = TraceableString(content, None)
            ts.origins = origins
            return ts
        return self.content[item]

    def get_origin(self, i):
        if i >= len(self):
            raise IndexError()
        for o in self.origins:
            if i in o:
                return o.origin, o.get_offset(i)
        return None

    def _populate_line_index(self):
        self._line_index = compute_row_index(self.content)

    def get_origin_of_line(self, line, col=0):
        if self._line_index is None:
            self._populate_line_index()
        i = self._line_index[line] + col
        return self.get_origin(i)

    def to_json(self):
        return {
            "content": self.content,
            "origins": [
                {
                    "begin": o.begin,
                    "end": o.end,
                    "origin": str(o.origin),
                    "offset": o.offset,
                }
                for o in self.origins
            ],
        }

    @staticmethod
    def from_json(data) -> "TraceableString":
        """
        Load a traceable string from a json.
        :param data: JSON as dictionary
        :return: The traceable string.
        """
        try:
            ts = TraceableString(data["content"], None)
            ts.origins = [
                OriginOfRange(
                    int(o["begin"]), int(o["end"]), o["origin"], int(o["offset"])
                )
                for o in data["origins"]
            ]
        except (KeyError, TypeError) as e:
            msg = f"Data not compatible: {e}"
            raise ValueError(msg)
        return ts

    def __add__(self, other):
        ts = TraceableString(self.content + other.content, None)
        ts.origins = self.origins + [o.move(len(self)) for o in other.origins]
        return ts

    def __eq__(self, other):
        if not isinstance(other, TraceableString):
            return False
        return self.content == other.content and self.origins == other.origins

    def __repr__(self):
        return f"TraceableString({self.content}, {self.origins})"



from setuptools import find_packages, setup


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="flachtex",
    version="0.3.15",
    description="A traceable LaTeX flattener.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    keywords="LaTeX flatten",
    url="https://github.com/d-krupke/flachtex",
    author="Dominik Krupke",
    author_email="krupke@ibr.cs.tu-bs.de",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=[],
    entry_points={
        "console_scripts": ["flachtex=flachtex.__main__:main"],
    },
    python_requires=">=3.8",
    include_package_data=True,
    zip_safe=False,
)



"""
Some parts should be skipped for the flattening. Here you can specify corresponding
rules.
"""

import abc
import re
import typing

from ..command_finder import CommandFinder
from ..traceable_string import TraceableString
from ..utils import Range


class SkipRule(abc.ABC):
    """
    A rule that defines, which parts should be skipped.
    """

    @abc.abstractmethod
    def find_all(self, content) -> typing.Iterable[Range]:
        pass


class RegexSkipRule(SkipRule):
    def __init__(self, regex: str):
        self.regex = re.compile(regex, re.MULTILINE | re.DOTALL)

    def find_all(self, content) -> typing.Iterable[Range]:
        for match in self.regex.finditer(content):
            yield self.determine_skip(match)

    @abc.abstractmethod
    def determine_skip(self, match: re.Match) -> Range:
        pass


class TodonotesRule(SkipRule):
    """
    Skips all the comments added with \\usepackage{todonotes}, or more concrete, all
    commands with \\todo[...]{...}.
    """

    def find_all(self, content) -> typing.Iterable[Range]:
        cf = CommandFinder()
        cf.add_command("todo", 1, 1)
        for match in cf.find_all(str(content)):
            yield Range(match.start, match.end)


class BasicSkipRule(RegexSkipRule):
    """
        Skips parts of the form
        ```
    %%FLACHTEX-SKIP-START
    ...
    %%FLACHTEX-SKIP-STOP
    """

    def __init__(self):
        super().__init__(
            r"(?P<skipped_part>(^\s*%%FLACHTEX-SKIP-START).*?(^\s*%%FLACHTEX-SKIP-STOP))"
        )

    def determine_skip(self, match: re.Match):
        span_to_be_skipped = Range(
            match.start("skipped_part"), match.end("skipped_part")
        )
        return span_to_be_skipped


def _find_skips(content, skip_rules):
    content = str(content)
    skips = []
    for rule in skip_rules:
        skips += list(rule.find_all(content))
    return skips


def _sort_and_check_ranges(skips) -> typing.Iterable[Range]:
    skips.sort()
    for i, e in enumerate(skips[:-1]):
        if e.intersects(skips[i + 1]):
            msg = "Intersecting skipped parts."
            raise ValueError(msg)
    return skips


def apply_skip_rules(
    content: TraceableString, skip_rules: typing.Iterable[SkipRule]
) -> TraceableString:
    """
    Apply a list of SkipRules to a content and return the content removed of all
    parts specified by the rules.
    :param content: The content to be filtered
    :param skip_rules: The rules to be applied
    :return:
    """
    skips = _find_skips(content, skip_rules)
    sorted_skips = _sort_and_check_ranges(skips)
    offset = 0
    for skip in sorted_skips:
        content = content[: skip.start + offset] + content[skip.end + offset :]
        offset -= len(skip)
    return content



import argparse
import json
import os

from .command_substitution import NewCommandSubstitution, find_new_commands
from .comments import remove_comments
from .preprocessor import Preprocessor
from .rules import ChangesRule, SubimportChangesRule, TodonotesRule


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="flachtex: Traceable LaTeX flattening."
    )
    parser.add_argument("--to_json", action="store_true", help="Return a json.")
    parser.add_argument("--comments", action="store_true", help="Remove comments.")
    parser.add_argument("--attach", action="store_true", help="Attach sources to json.")
    parser.add_argument(
        "--changes",
        action="store_true",
        help="Replace the commands of the changes package.",
    )
    parser.add_argument(
        "--changes_prefix",
        action="store_true",
        help="Use the prefix option in changes.",
    )
    parser.add_argument("--todos", action="store_true", help="Remove todo-notes.")
    parser.add_argument(
        "--newcommand",
        action="store_true",
        help="Automatically substitute custom commands.",
    )
    parser.add_argument("path", nargs=1, help="Path to main.tex")
    args = parser.parse_args()
    if not args.path:
        parser.print_help()
        exit(1)
    return args


def find_command_definitions(path) -> NewCommandSubstitution:
    """
    Parse the document once independently to extract new commands.
    :param path:
    :return:
    """
    preprocessor = Preprocessor(os.path.dirname(path))
    doc = preprocessor.expand_file(path)
    cmds = find_new_commands(doc)
    ncs = NewCommandSubstitution()
    for cmd in cmds:
        ncs.new_command(cmd)
    return ncs


def main():
    args = parse_arguments()
    file_path = args.path[0]
    preprocessor = Preprocessor(os.path.dirname(file_path))
    if args.todos:
        preprocessor.skip_rules.append(TodonotesRule())
    if args.changes:
        preprocessor.substitution_rules.append(ChangesRule(args.changes_prefix))
    if args.newcommand:
        preprocessor.substitution_rules.append(find_command_definitions(file_path))
    preprocessor.subimport_rules.append(SubimportChangesRule())
    doc = preprocessor.expand_file(file_path)

    if args.comments:
        doc = remove_comments(doc)
    if args.to_json:
        data = doc.to_json()
        data["sources"] = preprocessor.structure
        print(json.dumps(data))
    else:
        print(str(doc))


if __name__ == "__main__":
    main()



"""
Some parts of the LaTeX-document may need to be substituted.
The rules defined in this file allow such a handling.
"""

import abc
import logging
import typing

from flachtex.command_finder import CommandFinder, CommandMatch
from flachtex.traceable_string import TraceableString
from flachtex.utils import Range


class SubimportSubstitution(Range):
    def __init__(
        self, start: int, end: int, replacement_text: typing.Optional[TraceableString]
    ):
        super().__init__(start, end)
        assert not replacement_text or isinstance(replacement_text, TraceableString)
        self.replacement_text = replacement_text

    def __repr__(self):
        return f"REPLACEMENT[{self.start}:{self.end}]{{{self.replacement_text}}}"


class SubimportSubstitutionRule(abc.ABC):
    @abc.abstractmethod
    def find_all(
        self, content: TraceableString, subimport_path=None
    ) -> typing.Iterable[SubimportSubstitution]:
        pass


class SubimportChangesRule(SubimportSubstitutionRule):
    def __init__(self):
        self._includegraphics = ["includegraphics", 0]
        self._bibliography = ["bibliography", 0]

    def which(self, match: CommandMatch):
        members = [
            attr
            for attr in dir(self)
            if not callable(getattr(self, attr)) and not attr.startswith("__")
        ]
        members = members[1:]
        for m in members:
            if match.command == m[1:]:
                return vars(self)[m][1]

    def find_all(
        self, content: TraceableString, subimport_path=None
    ) -> typing.Iterable[SubimportSubstitution]:
        assert isinstance(content, TraceableString)
        cf = CommandFinder()
        cf.add_command(self._includegraphics[0], 1, 1)
        cf.add_command(self._bibliography[0], 1, 0)

        for match in cf.find_all(str(content)):
            pos_path = self.which(match)
            if (
                content.content[
                    match.parameters[pos_path][0] : match.parameters[pos_path][1]
                ][0:1]
                == "/"
            ):
                yield SubimportSubstitution(
                    match.parameters[pos_path][0],
                    match.parameters[pos_path][1],
                    TraceableString(
                        "./"
                        + subimport_path
                        + "/"
                        + content.content[
                            match.parameters[pos_path][0] : match.parameters[pos_path][
                                1
                            ]
                        ][1:],
                        content.origins,
                    ),
                )
            elif (
                content.content[
                    match.parameters[pos_path][0] : match.parameters[pos_path][1]
                ][0:2]
                == "./"
            ):
                yield SubimportSubstitution(
                    match.parameters[pos_path][0],
                    match.parameters[pos_path][1],
                    TraceableString(
                        "./"
                        + subimport_path
                        + "/"
                        + content.content[
                            match.parameters[pos_path][0] : match.parameters[pos_path][
                                1
                            ]
                        ][2:],
                        content.origins,
                    ),
                )
            else:
                yield SubimportSubstitution(
                    match.parameters[pos_path][0],
                    match.parameters[pos_path][1],
                    TraceableString(
                        "./"
                        + subimport_path
                        + "/"
                        + content.content[
                            match.parameters[pos_path][0] : (
                                match.parameters[pos_path][-1]
                            )
                        ],
                        content.origins,
                    ),
                )


def _sort_subimport_replacements(
    replacements: typing.List[SubimportSubstitution],
) -> typing.Iterable[SubimportSubstitution]:
    replacements.sort()
    if len(replacements) <= 1:
        return replacements
    replacements_ = []
    for i, e in enumerate(replacements):
        if e.intersects(replacements[(i + 1) % len(replacements)]):
            continue
        replacements_ += [e]
    return replacements_


def _find_subimport_substitutions(
    content: TraceableString,
    replacement_rules: typing.List[SubimportSubstitutionRule],
    subimport_path,
) -> typing.List[SubimportSubstitution]:
    replacements = []
    for rule in replacement_rules:
        replacements += list(rule.find_all(content, subimport_path))
    return replacements


def apply_subimport_substitution_rules(
    content: TraceableString,
    replacement_rules: typing.List[SubimportSubstitutionRule],
    subimport_path,
):
    replacements = _find_subimport_substitutions(
        content, replacement_rules, subimport_path
    )
    max_itererations = 10
    while replacements and max_itererations:
        replacements = _sort_subimport_replacements(replacements)
        offset = 0
        for replacement in replacements:
            if replacement.replacement_text:
                content = (
                    content[: replacement.start + offset]
                    + replacement.replacement_text
                    + content[replacement.end + offset :]
                )
                offset -= len(replacement) - len(replacement.replacement_text)
            else:
                content = (
                    content[: replacement.start + offset]
                    + content[replacement.end + offset :]
                )
                offset -= len(replacement)
        return content
    if max_itererations == 0:
        logging.getLogger("flachtex").warning(
            "Exceeded maximal subimport_replacement iterations."
        )
    return content
